/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Queso application class.
     */
    public class Application : Gtk.Application {

        /**
         * The “event_ocurred” signal.
         *
         * The ''::event_ocurred'' signal will be emitted when an event in the
         * application occurs.
         */
        public signal void event_ocurred(QuesoEvent event);

        /**
         * Whether the application is recording.
         */
        public bool is_recording { get { return camera.saving; } }

        /* The camera. */
        private Camera camera;
        /* The devices. */
        private Devices devices;
        /* The application window. */
        private ApplicationWindow window;

        /* Defines the initial state of the application. */
        construct {
            application_id = "mx.unam.Queso";
            devices = new Devices();
            camera = new Camera(devices.video_device,
                                devices.audio_device);
            devices.device_changed.connect(device_changed);
        }

        /**
         * Starts up the application.
         */
        public override void startup() {
            base.startup();

            var action = new GLib.SimpleAction("toggle-record", null);
            action.activate.connect(toggle_record);
            add_action(action);
            set_accels_for_action("app.toggle-record", new string[]{"<Ctrl>R"});

            action = new GLib.SimpleAction("info", null);
            action.activate.connect(info);
            add_action(action);
            set_accels_for_action("app.info", new string[]{"<Ctrl>I"});

            action = new GLib.SimpleAction("quit", null);
            action.activate.connect(quit);
            add_action(action);
            set_accels_for_action("app.quit", new string[]{"<Ctrl>Q"});
        }

        /**
         * Activates the application.
         */
        public override void activate() {
            if (window == null)
                window = new ApplicationWindow(this);
            present_and_start();
        }

        public void change_video_device(string name) {
            devices.update_video_device(name);
        }

        public void change_audio_device(string name) {
            devices.update_audio_device(name);
        }

        public void info() {
            camera.toggle_info();
        }

        /**
         * Quits the application.
         */
        public new void quit() {
            camera.clean();
            base.quit();
        }

        /* Toggles the recording. */
        private void toggle_record() {
            camera.toggle_save();
            event_ocurred(RECORDING_TOGGLED);
        }

        /* Adds the camera, presents the window and starts the camera. */
        private void present_and_start() {
            var videodev = devices.video_devices.first();
            var setup = videodev.setups.first();
            var w = setup.width;
            var h = setup.height;
            while (w > 1024) {
                w /= 2;
                h /= 2;
            }
            window.add_devices(devices);
            window.add_camera_widget(camera.widget, w, h);
            window.present();
            camera.start();
        }

        private void device_changed(DeviceType device_type) {
            switch (device_type) {
            case AUDIO:
                window.audio_device_changed(devices.audio_device);
                camera.audio_device_changed();
                break;
            case VIDEO:
                //window.video_device_changed(devices.video_device);
                //camera.video_device_changed();
                break;
            }
        }
    }
}
