/*
 * This file is part of queso.
 *
 * Copyright © 2020 Canek Peláez Valdés
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Author:
 *    Canek Peláez Valdés <canek@ciencias.unam.mx>
 */

namespace Queso {

    /**
     * Class for the Quesocamera.
     */
    public class Camera : GLib.Object {

        /* Null device. */
        private const string DEV_NULL = "/dev/null";

        /**
         * The widget.
         */
        public Gtk.Widget widget { public get; private set; }

        /**
         * Whether the camera is saving.
         */
        public bool saving { public get; private set; }

        /* The video and audio devices. */
        private VideoDevice videodev;
        private AudioDevice audiodev;
        private VideoSetup video_setup;
        private AudioSetup audio_setup;

        /* Pipeline and elements. */
        private Gst.Pipeline pipeline;
        private Gst.Bus bus;
        private Gst.Element vsrc;
        private Gst.Element asrc;
        private Gst.Element vfilter;
        private Gst.Element afilter;
        private Gst.Element parse;
        private Gst.Element aq1;
        private Gst.Element aq2;
        private Gst.Element vq1;
        private Gst.Element vq2;
        private Gst.Element tee;
        private Gst.Element dec;
        private Gst.Element ar;
        private Gst.Element ac;
        private Gst.Element aenc;
        private Gst.Element vc;
        private Gst.Element text;
        private Gst.Element info;
        private Gst.Element gtksink;
        private Gst.Element mp4mux;
        private Gst.Element filesink;

        /* Start/stop saving logic. */
        private Gst.Pad aq2_pad;
        private Gst.Pad vq1_pad;
        private Gst.Pad vq2_pad;
        private ulong aq2_pad_probe_id;
        private ulong vq2_pad_probe_id;
        private int buffer_count;
        private string temporary_fn;
        private string destination_fn;
        private bool video_ready;

        /* Timestamps. */
        private double start_saving_ts;
        private double last_saving_ts;

        private bool show_info;

        /**
         * Constructs a camera object for a video device.
         */
        public Camera(VideoDevice videodev, AudioDevice audiodev) {
            this.videodev = videodev;
            this.audiodev = audiodev;
            video_setup = videodev.setups.first();
            audio_setup = audiodev.setups.first();
            temporary_fn = random_filename();
            pipeline = new Gst.Pipeline("queso");

            asrc = audiodev.create_element("asrc");
            afilter = Gst.ElementFactory.make("capsfilter", "afilter");
            int rate = 44100;
            while (rate > audio_setup.max_rate)
                rate /= 2;
            int channels = audio_setup.max_rate > 1 ? 2 : 1;
            var fmt = audio_setup.formats.first();
            var acaps = new Gst.Caps.empty_simple(audio_setup.media_type);
            acaps.set_simple("format", typeof(string), fmt.format,
                             "rate", typeof(int), rate,
                             "channels", typeof(int), channels);
            afilter.set("caps", acaps);
            aq1 = Gst.ElementFactory.make("queue", "aq1");
            aq2 = Gst.ElementFactory.make("queue", "aq2");
            aq2.set("max-size-time", 3 * Gst.SECOND);
            aq2.set("max-size-bytes", 0);
            aq2.set("max-size-buffers", 0);
            aq2.set("leaky", 2);
            ar = Gst.ElementFactory.make("audiorate", "ar");
            ac = Gst.ElementFactory.make("audioconvert", "ac");
            aenc = Gst.ElementFactory.make("avenc_aac", "aenc");
            vsrc = videodev.create_element("vsrc");
            vfilter = Gst.ElementFactory.make("capsfilter", "vfilter");
            var fr = video_setup.framerates.first();
            var vcaps = new Gst.Caps.empty_simple(video_setup.media_type);
            vcaps.set_simple("framerate", typeof(Gst.Fraction),
                             fr.numerator, fr.denominator,
                             "width", typeof(int), video_setup.width,
                             "height", typeof(int), video_setup.height);
            vfilter.set("caps", vcaps);
            parse = Gst.ElementFactory.make("h264parse", "parse");
            parse.set("config_interval", -1);
            tee = Gst.ElementFactory.make("tee", "tee");
            vq1 = Gst.ElementFactory.make("queue", "vq1");
            dec = Gst.ElementFactory.make("avdec_h264", "dec");
            vc = Gst.ElementFactory.make("videoconvert", "vc");
            text = Gst.ElementFactory.make("textoverlay", "text");
            text.set("font-desc", "sans 14");
            text.set("halignment", 2);
            text.set("valignment", 2);
            info = Gst.ElementFactory.make("textoverlay", "info");
            info.set("font-desc", "sans 6");
            info.set("halignment", 0);
            info.set("valignment", 2);
            info.set("line-alignment", 0);
            gtksink = Gst.ElementFactory.make("gtksink", "gtksink");
            vq2 = Gst.ElementFactory.make("queue", "vq2");
            vq2.set("max-size-time", 3 * Gst.SECOND);
            vq2.set("max-size-bytes", 0);
            vq2.set("max-size-buffers", 0);
            vq2.set("leaky", 2);
            mp4mux = Gst.ElementFactory.make("mpegtsmux", "mp4mux");
            filesink = Gst.ElementFactory.make("filesink", "filesink");
            filesink.set("async", false);
            filesink.set("location", temporary_fn);

            pipeline.add_many(vsrc, vfilter, parse, tee,
                              vq1, dec, vc, text, info, gtksink,
                              vq2, mp4mux, filesink,
                              asrc, afilter, aq1, ar, ac, aenc, aq2);
            vsrc.link_many(vfilter, parse, tee);
            tee.link_many(vq1, dec, vc, text, info, gtksink);
            tee.link_many(vq2, mp4mux, filesink);
            asrc.link_many(afilter, aq1, ar, ac, aenc, aq2, mp4mux);

            Gtk.Widget w;
            gtksink.get("widget", out w);
            widget = w;

            bus = pipeline.get_bus();

            vq1_pad = vq1.get_static_pad("src");
            vq2_pad = vq2.get_static_pad("src");
            aq2_pad = aq2.get_static_pad("src");

            aq2_pad_probe_id = aq2_pad.add_probe(
                Gst.PadProbeType.BLOCK | Gst.PadProbeType.BUFFER,
                saving_block_probe);
            vq2_pad_probe_id = vq2_pad.add_probe(
                Gst.PadProbeType.BLOCK | Gst.PadProbeType.BUFFER,
                saving_block_probe);

            bus.add_watch(GLib.Priority.DEFAULT, pipeline_bus_watch);
        }

        /**
         * Cleans the camera and stops the pipeline.
         */
        public void clean() {
            pipeline.send_event(new Gst.Event.eos());
            filesink.set_state(Gst.State.NULL);
            mp4mux.set_state(Gst.State.NULL);
            filesink.set("location", DEV_NULL);
            if (saving) {
                destination_fn = get_destination();
                move_file(temporary_fn, destination_fn);
                destination_fn = null;
            }
            GLib.File tmp = GLib.File.new_for_path(temporary_fn);
            try {
                tmp.delete();
            } catch (GLib.Error e) {
                /* Nothing to do. */
            }
        }

        /**
         * Toggles saving the camera video.
         */
        public void toggle_save() {
            saving = !saving;
            if (saving) {
                start_saving();
            } else {
                stop_saving();
                text.set("text", "");
            }
        }

        /**
         * Starts the camera.
         */
        public void start() {
            GLib.Idle.add(start_camera);
        }

        /**
         * Stops the camera.
         */
        public void stop() {
            GLib.Idle.add(stop_camera);
        }

        /* Starts saving. */
        private void start_saving() {
            video_ready = false;
            buffer_count = 0;

            vq2_pad.add_probe(Gst.PadProbeType.BUFFER, wait_keyframe_probe);
            vq2_pad.remove_probe(vq2_pad_probe_id);
            vq2_pad_probe_id = 0;

            aq2_pad.add_probe(Gst.PadProbeType.BUFFER, wait_video_probe);
            aq2_pad.remove_probe(aq2_pad_probe_id);
            aq2_pad_probe_id = 0;
        }

        /* Stops saving. */
        private void stop_saving() {
            Gst.Debug.bin_to_dot_file(pipeline,
                                      Gst.DebugGraphDetails.ALL,
                                      "queso-pipeline");
            aq2_pad_probe_id = aq2_pad.add_probe(
                Gst.PadProbeType.BLOCK | Gst.PadProbeType.BUFFER,
                saving_block_probe);
            vq2_pad_probe_id = vq2_pad.add_probe(
                Gst.PadProbeType.BLOCK | Gst.PadProbeType.BUFFER,
                saving_block_probe);
            new Thread<void*>("eos-push-thread", push_eos_thread);
        }

        /* Starts the camera asynchronously. */
        private bool start_camera() {
            return change_pipeline_state(Gst.State.PLAYING);
        }

        /* Stops the camera asynchronously. */
        private bool stop_camera() {
            return change_pipeline_state(Gst.State.PAUSED);
        }

        /* Pipeline bus watch. */
        private bool
        pipeline_bus_watch(Gst.Bus bus, Gst.Message message) {
            switch (message.type) {
            case Gst.MessageType.ELEMENT:
                unowned Gst.Structure structure = message.get_structure();
                if (!structure.has_name("GstBinForwarded"))
                    return GLib.Source.CONTINUE;
                Gst.Message forward_message = null;
                structure.get("message", typeof(Gst.Message),
                              out forward_message);
                if (forward_message.type != Gst.MessageType.EOS)
                    return GLib.Source.CONTINUE;
                filesink.set_state(Gst.State.NULL);
                mp4mux.set_state(Gst.State.NULL);
                var old_fn = temporary_fn;
                temporary_fn = random_filename();
                filesink.set("location", temporary_fn);
                filesink.set_state(Gst.State.PLAYING);
                mp4mux.set_state(Gst.State.PLAYING);
                destination_fn = get_destination();
                move_file(old_fn, destination_fn);
                destination_fn = null;
                return GLib.Source.CONTINUE;
            default:
                return GLib.Source.CONTINUE;
            }
        }

        /* Saving block probe. */
        private Gst.PadProbeReturn
        saving_block_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            return Gst.PadProbeReturn.OK;
        }

        /* Wait keyframe probe. */
        private Gst.PadProbeReturn
        wait_keyframe_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            Gst.Buffer buffer = (Gst.Buffer)info.data;

            var flags = buffer.get_flags();
            if ((flags & Gst.BufferFlags.DELTA_UNIT) == 0) {
                start_saving_ts = last_saving_ts = Timestamp.get_timestamp();
                text.set("text", "•REC 00:00:00");
                GLib.Idle.add(update_counter);
                video_ready = true;
                return Gst.PadProbeReturn.REMOVE;
            } else {
                return Gst.PadProbeReturn.DROP;
            }
        }

        /* Drop one buffer probe. */
        private Gst.PadProbeReturn
        wait_video_probe(Gst.Pad pad, Gst.PadProbeInfo info) {
            if (!video_ready)
                return Gst.PadProbeReturn.DROP;
            return Gst.PadProbeReturn.REMOVE;
        }

        /* Pushes the EOS event. */
        private void*
        push_eos_thread() {
            var vpeer = vq2_pad.get_peer();
            var apeer = aq2_pad.get_peer();
            pipeline.message_forward = true;
            vpeer.send_event(new Gst.Event.eos());
            apeer.send_event(new Gst.Event.eos());
            return null;
        }

        /* Update counter idle callback. */
        private bool update_counter() {
            if (!saving) {
                text.set("text", "");
                return GLib.Source.REMOVE;
            }
            Gst.State state = Gst.State.NULL;
            Gst.State pending;
            pipeline.get_state(out state, out pending, 100);
            if (state != Gst.State.PLAYING)
                return GLib.Source.CONTINUE;
            double ts = Timestamp.get_timestamp();
            if (ts - last_saving_ts < 1.0)
                return GLib.Source.CONTINUE;
            last_saving_ts = ts;
            string sts = Timestamp.to_string(ts - start_saving_ts);
            text.set("text", "•REC " + sts);

            return GLib.Source.CONTINUE;
        }

        /* Changes the pipeline state, returning until it's changed. */
        private bool change_pipeline_state(Gst.State new_state) {
            pipeline.set_state(new_state);
            Gst.State state = Gst.State.NULL;
            Gst.State pending;

            Gst.StateChangeReturn r;

            do {
                r = pipeline.get_state(out state, out pending, 100);
                if (r == Gst.StateChangeReturn.FAILURE) {
                    return false;
                }
                Thread.usleep(10);
            } while (state != new_state);

            return true;
        }

        /* Gets the final destination for the video. */
        private string get_destination() {
            var now = new GLib.DateTime.now_local();
            string sts = "%04d-%02d-%02d %02d:%02d:%02d".printf(
                now.get_year(), now.get_month(), now.get_day_of_month(),
                now.get_hour(), now.get_minute(), now.get_second());
            var video_dir = GLib.Environment.get_user_special_dir(
                GLib.UserDirectory.VIDEOS);
            return string.join(GLib.Path.DIR_SEPARATOR_S, video_dir,
                               "Queso-%s.mp4".printf(sts));
        }

        /* Gets a random filename. */
        private string random_filename() {
            int ts = (int)Timestamp.get_timestamp();
            return string.join(GLib.Path.DIR_SEPARATOR_S,
                               GLib.Environment.get_tmp_dir(),
                               "queso-%08X.mp4".printf(ts));
        }

        /* Moves a file. */
        private void move_file(string src_path, string dest_path) {
            GLib.File src = GLib.File.new_for_path(src_path);
            GLib.File dest = GLib.File.new_for_path(dest_path);

            try {
                src.move(dest, GLib.FileCopyFlags.NONE, null);
            } catch (GLib.Error e) {
                GLib.warning("Could not move “%s” to “%s”: %s",
                             src_path, dest_path, e.message);
            }
        }

        private string get_info() {
            var r = new StringBuilder();
            r.append("Video source: ");
            r.append(videodev.name);
            r.append("\n");
            r.append("Audio source: ");
            r.append(audiodev.name);
            return r.str;
        }

        public void toggle_info() {
            show_info = !show_info;
            if (show_info) {
                info.set("text", get_info());
            } else {
                info.set("text", "");
            }
        }

        public void audio_device_changed() {
            
        }

        public void video_device_changed() {
            
        }
    }
}
